# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import os
from datetime import datetime

import fakeredis
import pytest

from integration.integration_test_utils import (
    V3ioHeaders,
    _generate_table_name,
    create_temp_kv,
    create_temp_redis_kv,
    drivers_list,
    get_redis_client,
    recursive_delete,
    remove_redis_table,
    remove_sql_tables,
)
from storey import V3ioDriver
from storey.redis_driver import RedisDriver
from storey.sql_driver import SQLDriver

SQLITE_DB = "sqlite:///test.db"


@pytest.fixture(params=drivers_list)
def setup_teardown_test(request):
    # Setup
    if request.param == "SQLDriver" and request.fspath.basename != "test_flow_integration.py":
        pytest.skip("SQLDriver test only in test_flow_integration")
    test_context = ContextForTests(request.param, table_name=_generate_table_name())

    # Test runs
    yield test_context

    # Teardown
    if test_context.driver_name == "V3ioDriver":
        asyncio.run(recursive_delete(test_context.table_name, V3ioHeaders()))
    elif test_context.driver_name == "RedisDriver":
        remove_redis_table(test_context.table_name)
    elif test_context.driver_name == "SQLDriver":
        remove_sql_tables()
    else:
        raise ValueError(f'Unsupported driver name "{test_context.driver_name}"')


@pytest.fixture(params=drivers_list)
def setup_kv_teardown_test(request):
    # Setup
    test_context = ContextForTests(request.param, table_name=_generate_table_name())

    if test_context.driver_name == "V3ioDriver":
        asyncio.run(create_temp_kv(test_context.table_name))
    elif test_context.driver_name == "RedisDriver":
        create_temp_redis_kv(test_context)
    elif test_context.driver_name == "SQLDriver":
        pytest.skip(msg="test not relevant for SQLDriver")
    else:
        raise ValueError(f'Unsupported driver name "{test_context.driver_name}"')

    # Test runs
    yield test_context

    # Teardown
    if test_context.driver_name == "V3ioDriver":
        asyncio.run(recursive_delete(test_context.table_name, V3ioHeaders()))
    elif test_context.driver_name == "RedisDriver":
        remove_redis_table(test_context.table_name)
    else:
        raise ValueError(f'Unsupported driver name "{test_context.driver_name}"')


@pytest.fixture()
def assign_stream_teardown_test():
    # Setup
    stream_path = _generate_table_name("bigdata/storey_ci/stream_test")

    # Test runs
    yield stream_path

    # Teardown
    asyncio.run(recursive_delete(stream_path, V3ioHeaders()))


# Can't call it TestContext because then pytest tries to run it as if it were a test suite
class ContextForTests:
    def __init__(self, driver_name: str, table_name: str):
        self._driver_name = driver_name
        self._table_name = table_name
        # sqlite cant save time zone
        self.test_base_time = (
            datetime.fromisoformat("2020-07-21T21:40:00+00:00")
            if driver_name != "SQLDriver"
            else datetime.fromisoformat("2020-07-21T21:40:00")
        )

        self._redis_fake_server = None
        if driver_name == "RedisDriver":
            redis_url = os.environ.get("MLRUN_REDIS_URL")
            if not redis_url:
                # if we are using fakeredis, create fake-server to support tests involving multiple clients
                self._redis_fake_server = fakeredis.FakeServer()
        if driver_name == "SQLDriver":
            self._sql_db_path = SQLITE_DB
            self._sql_table_name = table_name.split("/")[-2]
            self._table_name = f"{SQLITE_DB}/{self._sql_table_name}"

    @property
    def table_name(self):
        return self._table_name

    @property
    def redis_fake_server(self):
        return self._redis_fake_server

    @property
    def driver_name(self):
        return self._driver_name

    @property
    def sql_db_path(self):
        return self._sql_db_path

    class AggregationlessV3ioDriver(V3ioDriver):
        def supports_aggregations(self):
            return False

    class AggregationlessRedisDriver(RedisDriver):
        def supports_aggregations(self):
            return False

    def driver(self, *args, primary_key=None, is_aggregationless_driver=False, time_fields=None, **kwargs):
        if self.driver_name == "V3ioDriver":
            v3io_driver_class = ContextForTests.AggregationlessV3ioDriver if is_aggregationless_driver else V3ioDriver
            return v3io_driver_class(*args, **kwargs)
        elif self.driver_name == "RedisDriver":
            redis_driver_class = (
                ContextForTests.AggregationlessRedisDriver if is_aggregationless_driver else RedisDriver
            )
            return redis_driver_class(
                *args,
                redis_client=get_redis_client(self.redis_fake_server),
                key_prefix="storey-test:",
                **kwargs,
            )
        elif self.driver_name == "SQLDriver":
            if is_aggregationless_driver:
                sql_driver_class = SQLDriver
                return sql_driver_class(db_path=SQLITE_DB, primary_key=primary_key, time_fields=time_fields)
            else:
                pytest.skip("SQLDriver does not support aggregation")
        else:
            driver_name = self.driver_name
            raise ValueError(f'Unsupported driver name "{driver_name}"')



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime

from fsspec.implementations.local import LocalFileSystem

from storey.utils import find_filters, get_remaining_path, url_to_file_system


def test_get_path_utils():
    url = "wasbs://mycontainer@myaccount.blob.core.windows.net/path/to/object.csv"
    schema, path = get_remaining_path(url)
    assert path == "mycontainer/path/to/object.csv"
    assert schema == "wasbs"


def test_ds_get_path_utils():
    url = "ds://:file@profile/path/to/object.csv"
    fs, path = url_to_file_system(url, "")
    assert path == "/path/to/object.csv"
    assert isinstance(fs, LocalFileSystem)


def test_find_filters():
    filters = []
    find_filters([], datetime.datetime.min, datetime.datetime.max, filters, "time")
    assert filters == [[("time", ">", datetime.datetime.min), ("time", "<=", datetime.datetime.max)]]
    filters = []
    find_filters([], None, datetime.datetime.max, filters, "time")
    assert filters == [[("time", "<=", datetime.datetime.max)]]



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Optional

import pandas as pd

from .flow import Flow, _termination_obj


class ReduceToDataFrame(Flow):
    """Builds a pandas DataFrame from events and returns that DataFrame on flow termination.

    :param index: Name of the column to be used as index. Optional. If not set, DataFrame will be range indexed.
    :param columns: List of column names to be passed as-is to the DataFrame constructor. Optional.
    :param insert_key_column_as: Name of the column to be inserted for event keys. Optional.
        If not set, event keys will not be inserted into the DataFrame.
    :param insert_id_column_as: Name of the column to be inserted for event IDs. Optional.
        If not set, event IDs will not be inserted into the DataFrame.

    for additional params, see documentation of  :class:`storey.flow.Flow`

    """

    def __init__(
        self,
        index: Optional[str] = None,
        columns: Optional[List[str]] = None,
        insert_key_column_as: Optional[str] = None,
        insert_processing_time_column_as: Optional[str] = None,
        insert_id_column_as: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._index = index
        self._columns = columns
        self._insert_key_column_as = insert_key_column_as
        self._insert_processing_time_column_as = insert_processing_time_column_as
        self._insert_id_column_as = insert_id_column_as

    def _init(self):
        super()._init()
        self._key_column = []
        self._processing_time_column = []
        self._id_column = []
        self._data = []

    def to(self, outlet):
        """Pipe this step to next one. Throws exception since illegal"""
        raise ValueError("ToDataFrame is a terminal step. It cannot be piped further.")

    async def _do(self, event):
        if event is _termination_obj:
            df = pd.DataFrame(self._data, columns=self._columns)
            if not df.empty:
                if self._insert_key_column_as:
                    df[self._insert_key_column_as] = pd.DataFrame(self._key_column)
                if self._insert_processing_time_column_as:
                    df[self._insert_processing_time_column_as] = self._processing_time_column
                if self._insert_id_column_as:
                    df[self._insert_id_column_as] = self._id_column
                if self._index:
                    df.set_index(self._index, inplace=True)
            return df
        else:
            body = event.body
            if isinstance(body, dict) or isinstance(body, list):
                self._data.append(body)
                if self._insert_key_column_as:
                    self._key_column.append(event.key)
                if self._insert_processing_time_column_as:
                    self._processing_time_column.append(event.processing_time)
                if self._insert_id_column_as:
                    self._id_column.append(event.id)
            else:
                raise ValueError(f"ToDataFrame step only supports input of type dictionary or list, not {type(body)}")


class ToDataFrame(Flow):
    """Create pandas data frame from events. Can appear in the middle of the flow, as opposed to ReduceToDataFrame

    :param index: Name of the column to be used as index. Optional. If not set, DataFrame will be range indexed.
    :param columns: List of column names to be passed as-is to the DataFrame constructor. Optional.

    for additional params, see documentation of  :class:`storey.flow.Flow`
    """

    def __init__(self, index: Optional[str] = None, columns: Optional[List[str]] = None, **kwargs):
        super().__init__(**kwargs)
        self._index = index
        self._columns = columns

    async def _do(self, event):
        if event is _termination_obj:
            return await self._do_downstream(_termination_obj)
        else:
            df = pd.DataFrame(event.body, columns=self._columns)
            if self._index:
                df.set_index(self._index, inplace=True)
            event.body = df
            return await self._do_downstream(event)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import os
import uuid

import pandas as pd
import pytest

from storey import (
    AsyncEmitSource,
    CSVSource,
    CSVTarget,
    FlatMap,
    Map,
    ParquetTarget,
    Reduce,
    SyncEmitSource,
    build_flow,
)

from .integration_test_utils import _generate_table_name

has_azure_credentials = (
    os.getenv("AZURE_ACCOUNT_NAME") and os.getenv("AZURE_ACCOUNT_KEY") and os.getenv("AZURE_BLOB_STORE")
)
if has_azure_credentials:
    storage_options = {
        "account_name": os.getenv("AZURE_ACCOUNT_NAME"),
        "account_key": os.getenv("AZURE_ACCOUNT_KEY"),
    }
    from adlfs import AzureBlobFileSystem


@pytest.fixture()
def azure_create_csv():
    # Setup
    azure_blob = os.getenv("AZURE_BLOB_STORE")
    file_path = _generate_table_name(f"{azure_blob}/az_storey")

    _write_test_csv(file_path)

    # Test runs
    yield file_path

    # Teardown
    _delete_file(file_path)


@pytest.fixture()
def azure_teardown_file():
    # Setup
    azure_blob = os.getenv("AZURE_BLOB_STORE")
    file_path = _generate_table_name(f"{azure_blob}/az_storey")

    # Test runs
    yield file_path

    # Teardown
    _delete_file(file_path)


@pytest.fixture()
def azure_setup_teardown_test():
    # Setup
    table_name = _generate_table_name(f'{os.getenv("AZURE_BLOB_STORE")}/test')

    # Test runs
    yield table_name

    # Teardown
    azure_recursive_delete(table_name)


def _write_test_csv(file_path):
    az_fs = AzureBlobFileSystem(**storage_options)
    data = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    with az_fs.open(file_path, "w") as f:
        f.write(data)


def _delete_file(path):
    az_fs = AzureBlobFileSystem(**storage_options)
    az_fs.delete(path)


def azure_recursive_delete(path):
    az_fs = AzureBlobFileSystem(**storage_options)
    az_fs.rm(path, True)


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_csv_reader_from_azure(azure_create_csv):
    controller = build_flow(
        [
            CSVSource(
                f"az:///{azure_create_csv}",
                storage_options=storage_options,
            ),
            FlatMap(lambda x: x),
            Map(lambda x: int(x)),
            Reduce(0, lambda acc, x: acc + x),
        ]
    ).run()

    termination_result = controller.await_termination()
    assert termination_result == 495


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_csv_reader_from_azure_error_on_file_not_found():
    with pytest.raises(FileNotFoundError):
        controller = build_flow(
            [
                CSVSource(
                    f'az:///{os.getenv("AZURE_BLOB_STORE")}/idontexist.csv',
                    storage_options=storage_options,
                ),
            ]
        ).run()
        controller.await_termination()


async def async_test_write_csv_to_azure(azure_teardown_csv):
    controller = build_flow(
        [
            AsyncEmitSource(),
            CSVTarget(
                f"az:///{azure_teardown_csv}",
                columns=["n", "n*10"],
                header=True,
                storage_options=storage_options,
            ),
        ]
    ).run()

    for i in range(10):
        await controller.emit([i, 10 * i])

    await controller.terminate()
    await controller.await_termination()

    actual = AzureBlobFileSystem(**storage_options).open(azure_teardown_csv).read()

    expected = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert actual.decode("utf-8") == expected


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_csv_to_azure(azure_teardown_file):
    asyncio.run(async_test_write_csv_to_azure(azure_teardown_file))


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_csv_with_dict_to_azure(azure_teardown_file):
    file_path = f"az:///{azure_teardown_file}"
    controller = build_flow(
        [
            SyncEmitSource(),
            CSVTarget(
                file_path,
                columns=["n", "n*10"],
                header=True,
                storage_options=storage_options,
            ),
        ]
    ).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i})

    controller.terminate()
    controller.await_termination()

    actual = AzureBlobFileSystem(**storage_options).open(azure_teardown_file).read()
    expected = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert actual.decode("utf-8") == expected


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_csv_infer_columns_without_header_to_azure(azure_teardown_file):
    file_path = f"az:///{azure_teardown_file}"
    controller = build_flow([SyncEmitSource(), CSVTarget(file_path, storage_options=storage_options)]).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i})

    controller.terminate()
    controller.await_termination()

    actual = AzureBlobFileSystem(**storage_options).open(azure_teardown_file).read()
    expected = "0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert actual.decode("utf-8") == expected


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_csv_from_lists_with_metadata_and_column_pruning_to_azure(
    azure_teardown_file,
):
    file_path = f"az:///{azure_teardown_file}"
    controller = build_flow(
        [
            SyncEmitSource(),
            CSVTarget(
                file_path,
                columns=["event_key=$key", "n*10"],
                header=True,
                storage_options=storage_options,
            ),
        ]
    ).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i}, key=f"key{i}")

    controller.terminate()
    controller.await_termination()

    actual = AzureBlobFileSystem(**storage_options).open(azure_teardown_file).read()
    expected = (
        "event_key,n*10\n"
        "key0,0\n"
        "key1,10\n"
        "key2,20\n"
        "key3,30\n"
        "key4,40\n"
        "key5,50\n"
        "key6,60\n"
        "key7,70\n"
        "key8,80\n"
        "key9,90\n"
    )
    assert actual.decode("utf-8") == expected


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_to_parquet_to_azure(azure_setup_teardown_test):
    out_dir = f"az:///{azure_setup_teardown_test}"
    columns = ["my_int", "my_string"]
    controller = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(
                out_dir,
                partition_cols="my_int",
                columns=columns,
                max_events=1,
                storage_options=storage_options,
            ),
        ]
    ).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"])
        expected.append([i, f"this is {i}"])
    expected = pd.DataFrame(expected, columns=columns, dtype="int32")
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_dir, columns=columns, storage_options=storage_options)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_to_parquet_to_azure_single_file_on_termination(
    azure_setup_teardown_test,
):
    out_file = f"az:///{azure_setup_teardown_test}/out.parquet"
    columns = ["my_int", "my_string"]
    controller = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(out_file, columns=columns, storage_options=storage_options),
        ]
    ).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"])
        expected.append([i, f"this is {i}"])
    expected = pd.DataFrame(expected, columns=columns, dtype="int64")
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_file, columns=columns, storage_options=storage_options)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"


@pytest.mark.skipif(not has_azure_credentials, reason="No azure credentials found")
def test_write_to_parquet_to_azure_with_indices(azure_setup_teardown_test):
    out_file = f"az:///{azure_setup_teardown_test}/test_write_to_parquet_with_indices{uuid.uuid4().hex}.parquet"
    controller = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(
                out_file,
                index_cols="event_key=$key",
                columns=["my_int", "my_string"],
                storage_options=storage_options,
            ),
        ]
    ).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"], key=f"key{i}")
        expected.append([f"key{i}", i, f"this is {i}"])
    columns = ["event_key", "my_int", "my_string"]
    expected = pd.DataFrame(expected, columns=columns, dtype="int64")
    expected.set_index(["event_key"], inplace=True)
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_file, columns=columns, storage_options=storage_options)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import math

_aggrTypeNone = 0
_aggrTypeCount = 1
_aggrTypeSum = 2
_aggrTypeSqr = 4
_aggrTypeMax = 8
_aggrTypeMin = 16
_aggrTypeLast = 32
_aggrTypeFirst = 64

# Derived aggregates
_aggrTypeAvg = _aggrTypeCount | _aggrTypeSum
_aggrTypeRate = _aggrTypeLast | 0x8000
_aggrTypeStddev = _aggrTypeCount | _aggrTypeSum | _aggrTypeSqr
_aggrTypeStdvar = _aggrTypeCount | _aggrTypeSum | _aggrTypeSqr | 0x8000
_aggrTypeAll = 0xFFFF

_raw_aggregates = [
    _aggrTypeCount,
    _aggrTypeSum,
    _aggrTypeSqr,
    _aggrTypeMax,
    _aggrTypeMin,
    _aggrTypeLast,
]
_raw_aggregates_by_name = {
    "count": _aggrTypeCount,
    "sum": _aggrTypeSum,
    "sqr": _aggrTypeSqr,
    "max": _aggrTypeMax,
    "min": _aggrTypeMin,
    "first": _aggrTypeFirst,
    "last": _aggrTypeLast,
}
_all_aggregates_by_name = {
    "count": _aggrTypeCount,
    "sum": _aggrTypeSum,
    "sqr": _aggrTypeSqr,
    "max": _aggrTypeMax,
    "min": _aggrTypeMin,
    "first": _aggrTypeFirst,
    "last": _aggrTypeLast,
    "avg": _aggrTypeAvg,
    "stdvar": _aggrTypeStdvar,
    "stddev": _aggrTypeStddev,
}
_all_aggregates_to_name = {
    _aggrTypeCount: "count",
    _aggrTypeSum: "sum",
    _aggrTypeSqr: "sqr",
    _aggrTypeMax: "max",
    _aggrTypeMin: "min",
    _aggrTypeFirst: "first",
    _aggrTypeLast: "last",
    _aggrTypeAvg: "avg",
    _aggrTypeStdvar: "stdvar",
    _aggrTypeStddev: "stddev",
}


def is_raw_aggregate(aggregate):
    return aggregate in _raw_aggregates_by_name


def _avg(args):
    count = args[0]
    sum = args[1]
    if count == 0:
        return math.nan
    return sum / count


def _stddev(args):
    count = args[0]
    if count == 0 or count == 1:
        return math.nan
    sum = args[1]
    sqr = args[2]

    return math.sqrt((count * sqr - sum * sum) / (count * (count - 1)))


def _stdvar(args):
    count = args[0]
    if count == 0 or count == 1:
        return math.nan
    sum = args[1]
    sqr = args[2]
    return (count * sqr - sum * sum) / (count * (count - 1))


def get_virtual_aggregation_func(aggregation):
    if aggregation == "avg":
        return _avg
    if aggregation == "stdvar":
        return _stdvar
    if aggregation == "stddev":
        return _stddev

    raise TypeError(f'"{aggregation}" aggregator is not defined')


def get_implied_aggregates(aggregate):
    aggrs = []
    aggr_bits = _all_aggregates_by_name[aggregate]
    for raw_aggr in _raw_aggregates:
        if aggr_bits & raw_aggr == raw_aggr:
            aggrs.append(_all_aggregates_to_name[raw_aggr])
    return aggrs


def get_all_raw_aggregates_with_hidden(aggregates):
    raw_aggregates = {}

    for aggregate in aggregates:
        if is_raw_aggregate(aggregate):
            raw_aggregates[aggregate] = False
        else:
            for dependant_aggr in get_implied_aggregates(aggregate):
                if dependant_aggr not in raw_aggregates:
                    raw_aggregates[dependant_aggr] = True

    return raw_aggregates


def get_all_raw_aggregates(aggregates):
    return set(get_all_raw_aggregates_with_hidden(aggregates).keys())


def is_aggregation_name(name: str):
    return name in _all_aggregates_by_name
